@@ -88,7 +88,12 @@ from hypothesis.internal.reflection import (
     nicerepr,
     proxies,
 )
-from hypothesis.reporting import current_verbosity, report, verbose_report
+from hypothesis.reporting import (
+    current_verbosity,
+    report,
+    verbose_report,
+    with_reporter,
+)
 from hypothesis.searchstrategy.collections import TupleStrategy
 from hypothesis.searchstrategy.strategies import MappedSearchStrategy, SearchStrategy
 from hypothesis.statistics import note_engine_for_statistics
@@ -227,7 +232,15 @@ class WithRunner(MappedSearchStrategy):
         return "WithRunner(%r, runner=%r)" % (self.mapped_strategy, self.runner)
 
 
-def is_invalid_test(name, original_argspec, generator_arguments, generator_kwargs):
+def is_invalid_test(name, original_argspec, given_arguments, given_kwargs):
+    """Check the arguments to ``@given`` for basic usage constraints.
+
+    Most errors are not raised immediately; instead we return a dummy test
+    function that will raise the appropriate error if it is actually called.
+    When the user runs a subset of tests (e.g via ``pytest -k``), errors will
+    only be reported for tests that actually ran.
+    """
+
     def invalid(message):
         def wrapped_test(*arguments, **kwargs):
             raise InvalidArgument(message)
@@ -235,10 +248,10 @@ def is_invalid_test(name, original_argspec, generator_arguments, generator_kwarg
         wrapped_test.is_hypothesis_test = True
         return wrapped_test
 
-    if not (generator_arguments or generator_kwargs):
+    if not (given_arguments or given_kwargs):
         return invalid("given must be called with at least one argument")
 
-    if generator_arguments and any(
+    if given_arguments and any(
         [original_argspec.varargs, original_argspec.varkw, original_argspec.kwonlyargs]
     ):
         return invalid(
@@ -246,32 +259,32 @@ def is_invalid_test(name, original_argspec, generator_arguments, generator_kwarg
             "varkeywords, or keyword-only arguments"
         )
 
-    if len(generator_arguments) > len(original_argspec.args):
-        args = tuple(generator_arguments)
+    if len(given_arguments) > len(original_argspec.args):
+        args = tuple(given_arguments)
         return invalid(
             "Too many positional arguments for %s() were passed to @given "
             "- expected at most %d arguments, but got %d %r"
             % (name, len(original_argspec.args), len(args), args)
         )
 
-    if infer in generator_arguments:
+    if infer in given_arguments:
         return invalid(
             "infer was passed as a positional argument to @given, "
             "but may only be passed as a keyword argument"
         )
 
-    if generator_arguments and generator_kwargs:
+    if given_arguments and given_kwargs:
         return invalid("cannot mix positional and keyword arguments to @given")
     extra_kwargs = [
         k
-        for k in generator_kwargs
+        for k in given_kwargs
         if k not in original_argspec.args + original_argspec.kwonlyargs
     ]
     if extra_kwargs and not original_argspec.varkw:
         arg = extra_kwargs[0]
         return invalid(
             "%s() got an unexpected keyword argument %r, from `%s=%r` in @given"
-            % (name, arg, arg, generator_kwargs[arg])
+            % (name, arg, arg, given_kwargs[arg])
         )
     for a in original_argspec.args:
         if isinstance(a, list):  # pragma: no cover
@@ -281,9 +294,7 @@ def is_invalid_test(name, original_argspec, generator_arguments, generator_kwarg
             )
     if original_argspec.defaults or original_argspec.kwonlydefaults:
         return invalid("Cannot apply @given to a function with defaults.")
-    missing = [
-        repr(kw) for kw in original_argspec.kwonlyargs if kw not in generator_kwargs
-    ]
+    missing = [repr(kw) for kw in original_argspec.kwonlyargs if kw not in given_kwargs]
     if missing:
         return invalid(
             "Missing required kwarg{}: {}".format(
@@ -292,10 +303,37 @@ def is_invalid_test(name, original_argspec, generator_arguments, generator_kwarg
         )
 
 
-def execute_explicit_examples(
-    test_runner, test, wrapped_test, settings, arguments, kwargs
-):
-    original_argspec = getfullargspec(test)
+class ArtificialDataForExample(ConjectureData):
+    """Dummy object that pretends to be a ConjectureData object for the purposes of
+    drawing arguments for @example. Provides just enough of the ConjectureData API
+    to allow the test to run. Does not support any sort of interactive drawing,
+    but that's fine because you can't access that when all of your arguments are
+    provided by @example.
+    """
+
+    def __init__(self, kwargs):
+        self.__draws = 0
+        self.__kwargs = kwargs
+
+        def draw_bytes(data, n):
+            raise NotImplementedError()  # pragma: no cover
+
+        super(ArtificialDataForExample, self).__init__(
+            max_length=0, draw_bytes=draw_bytes
+        )
+
+    def draw(self, strategy):
+        assert self.__draws == 0
+        self.__draws += 1
+        # The main strategy for given is always a tuples strategy that returns
+        # first positional arguments then keyword arguments. When building this
+        # object already converted all positional arguments to keyword arguments,
+        # so this is the correct format to return.
+        return (), self.__kwargs
+
+
+def execute_explicit_examples(state, wrapped_test, arguments, kwargs):
+    original_argspec = getfullargspec(state.test)
 
     for example in reversed(getattr(wrapped_test, "hypothesis_explicit_examples", ())):
         example_kwargs = dict(original_argspec.kwonlydefaults or {})
@@ -311,27 +349,37 @@ def execute_explicit_examples(
             )
         else:
             example_kwargs.update(example.kwargs)
-        if Phase.explicit not in settings.phases:
+        if Phase.explicit not in state.settings.phases:
             continue
         example_kwargs.update(kwargs)
-        # Note: Test may mutate arguments and we can't rerun explicit
-        # examples, so we have to calculate the failure message at this
-        # point rather than than later.
-        example_string = "%s(%s)" % (
-            test.__name__,
-            arg_string(test, arguments, example_kwargs),
-        )
-        with local_settings(settings):
+
+        with local_settings(state.settings):
+            fragments_reported = []
+
+            def report_buffered():
+                for f in fragments_reported:
+                    report(f)
+                del fragments_reported[:]
+
             try:
-                with BuildContext(None) as b:
-                    verbose_report("Trying example: " + example_string)
-                    test_runner(None, lambda data: test(*arguments, **example_kwargs))
+                with with_reporter(fragments_reported.append):
+                    state.execute_once(
+                        ArtificialDataForExample(example_kwargs),
+                        is_final=True,
+                        print_example=True,
+                    )
             except BaseException:
-                report("Falsifying example: " + example_string)
-                for n in b.notes:
-                    report(n)
+                report_buffered()
                 raise
 
+            if current_verbosity() >= Verbosity.verbose:
+                prefix = "Falsifying example"
+                assert fragments_reported[0].startswith(prefix)
+                fragments_reported[0] = (
+                    "Trying example" + fragments_reported[0][len(prefix) :]
+                )
+                report_buffered()
+
 
 def get_random_for_wrapped_test(test, wrapped_test):
     settings = wrapped_test._hypothesis_internal_use_settings
@@ -350,14 +398,7 @@ def get_random_for_wrapped_test(test, wrapped_test):
 
 
 def process_arguments_to_given(
-    wrapped_test,
-    arguments,
-    kwargs,
-    generator_arguments,
-    generator_kwargs,
-    argspec,
-    test,
-    settings,
+    wrapped_test, arguments, kwargs, given_kwargs, argspec, test, settings,
 ):
     selfy = None
     arguments, kwargs = convert_positional_arguments(wrapped_test, arguments, kwargs)
@@ -385,9 +426,7 @@ def process_arguments_to_given(
     search_strategy = TupleStrategy(
         (
             st.just(arguments),
-            st.fixed_dictionaries(generator_kwargs).map(
-                lambda args: dict(args, **kwargs)
-            ),
+            st.fixed_dictionaries(given_kwargs).map(lambda args: dict(args, **kwargs)),
         )
     )
 
@@ -435,12 +474,10 @@ def failure_exceptions_to_catch():
     return tuple(exceptions)
 
 
-def new_given_argspec(original_argspec, generator_kwargs):
+def new_given_argspec(original_argspec, given_kwargs):
     """Make an updated argspec for the wrapped test."""
-    new_args = [a for a in original_argspec.args if a not in generator_kwargs]
-    new_kwonlyargs = [
-        a for a in original_argspec.kwonlyargs if a not in generator_kwargs
-    ]
+    new_args = [a for a in original_argspec.args if a not in given_kwargs]
+    new_kwonlyargs = [a for a in original_argspec.kwonlyargs if a not in given_kwargs]
     annots = {
         k: v
         for k, v in original_argspec.annotations.items()
@@ -470,9 +507,19 @@ class StateForActualGivenExecution(object):
         self.files_to_propagate = set()
         self.failed_normally = False
 
-        self.used_examples_from_database = False
+    def execute_once(
+        self, data, print_example=False, is_final=False, expected_failure=None
+    ):
+        """Run the test function once, using ``data`` as input.
 
-    def execute(self, data, print_example=False, is_final=False, expected_failure=None):
+        If the test raises an exception, it will propagate through to the
+        caller of this method. Depending on its type, this could represent
+        an ordinary test failure, or a fatal error, or a control exception.
+
+        If this method returns normally, the test might have passed, or
+        it might have placed ``data`` in an unsuccessful state and then
+        swallowed the corresponding control exception.
+        """
         text_repr = [None]
         if self.settings.deadline is None:
             test = self.test
@@ -498,9 +545,12 @@ class StateForActualGivenExecution(object):
                 return result
 
         def run(data):
+            # Set up dynamic context needed by a single test run.
             with local_settings(self.settings):
                 with deterministic_PRNG():
                     with BuildContext(data, is_final=is_final):
+
+                        # Generate all arguments to the test function.
                         args, kwargs = data.draw(self.search_strategy)
                         if expected_failure is not None:
                             text_repr[0] = arg_string(test, args, kwargs)
@@ -518,7 +568,12 @@ class StateForActualGivenExecution(object):
                             )
                         return test(*args, **kwargs)
 
+        # Run the test function once, via the executor hook.
+        # In most cases this will delegate straight to `run(data)`.
         result = self.test_runner(data, run)
+
+        # If a failure was expected, it should have been raised already, so
+        # instead raise an appropriate diagnostic error.
         if expected_failure is not None:
             exception, traceback = expected_failure
             if (
@@ -551,9 +606,16 @@ class StateForActualGivenExecution(object):
             )
         return result
 
-    def evaluate_test_data(self, data):
+    def _execute_once_for_engine(self, data):
+        """Wrapper around ``execute_once`` that intercepts test failure
+        exceptions and single-test control exceptions, and turns them into
+        appropriate method calls to `data` instead.
+
+        This allows the engine to assume that any exception other than
+        ``StopTest`` must be a fatal error, and should stop the entire engine.
+        """
         try:
-            result = self.execute(data)
+            result = self.execute_once(data)
             if result is not None:
                 fail_health_check(
                     self.settings,
@@ -565,21 +627,35 @@ class StateForActualGivenExecution(object):
                     HealthCheck.return_value,
                 )
         except UnsatisfiedAssumption:
+            # An "assume" check failed, so instead we inform the engine that
+            # this test run was invalid.
             data.mark_invalid()
+        except StopTest:
+            # The engine knows how to handle this control exception, so it's
+            # OK to re-raise it.
+            raise
         except (
             HypothesisDeprecationWarning,
             FailedHealthCheck,
-            StopTest,
         ) + skip_exceptions_to_reraise():
+            # These are fatal errors or control exceptions that should stop the
+            # engine, so we re-raise them.
             raise
         except failure_exceptions_to_catch() as e:
+            # If the error was raised by Hypothesis-internal code, re-raise it
+            # as a fatal error instead of treating it as a test failure.
             escalate_hypothesis_internal_error()
+
             if data.frozen:
                 # This can happen if an error occurred in a finally
                 # block somewhere, suppressing our original StopTest.
                 # We raise a new one here to resume normal operation.
                 raise StopTest(data.testcounter)
             else:
+                # The test failed by raising an exception, so we inform the
+                # engine that this test run was interesting. This is the normal
+                # path for test runs that fail.
+
                 tb = get_trimmed_traceback()
                 info = data.extra_information
                 info.__expected_traceback = "".join(
@@ -593,7 +669,10 @@ class StateForActualGivenExecution(object):
                 lineno = origin[1]
                 data.mark_interesting((type(e), filename, lineno))
 
-    def run(self):
+    def run_engine(self):
+        """Run the test function many times, on database input and generated
+        input, using the Conjecture engine.
+        """
         # Tell pytest to omit the body of this function from tracebacks
         __tracebackhide__ = True
         if global_force_seed is None:
@@ -601,19 +680,16 @@ class StateForActualGivenExecution(object):
         else:
             database_key = None
         runner = ConjectureRunner(
-            self.evaluate_test_data,
+            self._execute_once_for_engine,
             settings=self.settings,
             random=self.random,
             database_key=database_key,
         )
-        try:
-            runner.run()
-        finally:
-            self.used_examples_from_database = runner.used_examples_from_database
+        # Use the Conjecture engine to run the test function many times
+        # on different inputs.
+        runner.run()
         note_engine_for_statistics(runner)
 
-        self.used_examples_from_database = runner.used_examples_from_database
-
         if runner.call_count == 0:
             return
         if runner.interesting_examples:
@@ -632,8 +708,12 @@ class StateForActualGivenExecution(object):
         if not self.falsifying_examples:
             return
         elif not self.settings.report_multiple_bugs:
+            # Pretend that we only found one failure, by discarding the others.
             del self.falsifying_examples[:-1]
 
+        # The engine found one or more failures, so we need to reproduce and
+        # report them.
+
         self.failed_normally = True
 
         flaky = 0
@@ -645,7 +725,7 @@ class StateForActualGivenExecution(object):
             self.__was_flaky = False
             assert info.__expected_exception is not None
             try:
-                self.execute(
+                self.execute_once(
                     ran_example,
                     print_example=True,
                     is_final=True,
@@ -662,9 +742,15 @@ class StateForActualGivenExecution(object):
                 )
             except BaseException as e:
                 if len(self.falsifying_examples) <= 1:
+                    # There is only one failure, so we can report it by raising
+                    # it directly.
                     raise
+
+                # We are reporting multiple failures, so we need to manually
+                # print each exception's stack trace and information.
                 tb = get_trimmed_traceback()
                 report("".join(traceback.format_exception(type(e), e, tb)))
+
             finally:  # pragma: no cover
                 # This section is in fact entirely covered by the tests in
                 # test_reproduce_failure, but it seems to trigger a lovely set
@@ -751,8 +837,8 @@ class HypothesisHandle(object):
 
 
 def given(
-    *given_arguments,  # type: Union[SearchStrategy, InferType]
-    **given_kwargs  # type: Union[SearchStrategy, InferType]
+    *_given_arguments,  # type: Union[SearchStrategy, InferType]
+    **_given_kwargs  # type: Union[SearchStrategy, InferType]
 ):
     # type: (...) -> Callable[[Callable[..., None]], Callable[..., None]]
     """A decorator for turning a test function that accepts arguments into a
@@ -761,29 +847,37 @@ def given(
     This is the main entry point to Hypothesis.
     """
 
-    def run_test_with_generator(test):
+    def run_test_as_given(test):
         if inspect.isclass(test):
             # Provide a meaningful error to users, instead of exceptions from
             # internals that assume we're dealing with a function.
             raise InvalidArgument("@given cannot be applied to a class.")
-        generator_arguments = tuple(given_arguments)
-        generator_kwargs = dict(given_kwargs)
+        given_arguments = tuple(_given_arguments)
+        given_kwargs = dict(_given_kwargs)
 
         original_argspec = getfullargspec(test)
 
         check_invalid = is_invalid_test(
-            test.__name__, original_argspec, generator_arguments, generator_kwargs
+            test.__name__, original_argspec, given_arguments, given_kwargs
         )
 
+        # If the argument check found problems, return a dummy test function
+        # that will raise an error if it is actually called.
         if check_invalid is not None:
             return check_invalid
 
-        for name, strategy in zip(
-            reversed(original_argspec.args), reversed(generator_arguments)
-        ):
-            generator_kwargs[name] = strategy
+        # Because the argument check succeeded, we can convert @given's
+        # positional arguments into keyword arguments for simplicity.
+        if given_arguments:
+            assert not given_kwargs
+            for name, strategy in zip(
+                reversed(original_argspec.args), reversed(given_arguments)
+            ):
+                given_kwargs[name] = strategy
+        # These have been converted, so delete them to prevent accidental use.
+        del given_arguments
 
-        argspec = new_given_argspec(original_argspec, generator_kwargs)
+        argspec = new_given_argspec(original_argspec, given_kwargs)
 
         @impersonate(test)
         @define_function_signature(test.__name__, test.__doc__, argspec)
@@ -810,27 +904,22 @@ def given(
 
             random = get_random_for_wrapped_test(test, wrapped_test)
 
-            if infer in generator_kwargs.values():
+            # Use type information to convert "infer" arguments into appropriate
+            # strategies.
+            if infer in given_kwargs.values():
                 hints = get_type_hints(test)
             for name in [
-                name for name, value in generator_kwargs.items() if value is infer
+                name for name, value in given_kwargs.items() if value is infer
             ]:
                 if name not in hints:
                     raise InvalidArgument(
                         "passed %s=infer for %s, but %s has no type annotation"
                         % (name, test.__name__, name)
                     )
-                generator_kwargs[name] = st.from_type(hints[name])
+                given_kwargs[name] = st.from_type(hints[name])
 
             processed_args = process_arguments_to_given(
-                wrapped_test,
-                arguments,
-                kwargs,
-                generator_arguments,
-                generator_kwargs,
-                argspec,
-                test,
-                settings,
+                wrapped_test, arguments, kwargs, given_kwargs, argspec, test, settings,
             )
             arguments, kwargs, test_runner, search_strategy = processed_args
 
@@ -863,6 +952,9 @@ def given(
 
             reproduce_failure = wrapped_test._hypothesis_internal_use_reproduce_failure
 
+            # If there was a @reproduce_failure decorator, use it to reproduce
+            # the error (or complain that we couldn't). Either way, this will
+            # always raise some kind of error.
             if reproduce_failure is not None:
                 expected_version, failure = reproduce_failure
                 if expected_version != __version__:
@@ -876,7 +968,7 @@ def given(
                         % (expected_version, __version__)
                     )
                 try:
-                    state.execute(
+                    state.execute_once(
                         ConjectureData.for_buffer(decode_failure(failure)),
                         print_example=True,
                         is_final=True,
@@ -898,9 +990,14 @@ def given(
                         "generated?"
                     )
 
-            execute_explicit_examples(
-                test_runner, test, wrapped_test, settings, arguments, kwargs
-            )
+            # There was no @reproduce_failure, so start by running any explicit
+            # examples from @example decorators.
+
+            execute_explicit_examples(state, wrapped_test, arguments, kwargs)
+
+            # If there were any explicit examples, they all ran successfully.
+            # The next step is to use the Conjecture engine to run the test on
+            # many different inputs.
 
             if settings.max_examples <= 0:
                 return
@@ -915,12 +1012,16 @@ def given(
                     subTest = runner.subTest
                     try:
                         runner.subTest = fake_subTest
-                        state.run()
+                        state.run_engine()
                     finally:
                         runner.subTest = subTest
                 else:
-                    state.run()
+                    state.run_engine()
             except BaseException as e:
+                # The exception caught here should either be an actual test
+                # failure (or MultipleFailures), or some kind of fatal error
+                # that caused the engine to stop.
+
                 generated_seed = wrapped_test._hypothesis_internal_use_generated_seed
                 with local_settings(settings):
                     if not (state.failed_normally or generated_seed is None):
@@ -958,6 +1059,9 @@ def given(
                     )
                     raise the_error_hypothesis_found
 
+        # After having created the decorated test function, we need to copy
+        # over some attributes to make the switch as seamless as possible.
+
         for attrib in dir(test):
             if not (attrib.startswith("_") or hasattr(wrapped_test, attrib)):
                 setattr(wrapped_test, attrib, getattr(test, attrib))
@@ -977,7 +1081,7 @@ def given(
         wrapped_test.hypothesis = HypothesisHandle(test)
         return wrapped_test
 
-    return run_test_with_generator
+    return run_test_as_given
 
 
 def find(
@@ -1050,6 +1154,7 @@ def find(
     runner = ConjectureRunner(
         template_condition, settings=settings, random=random, database_key=database_key
     )
+
     runner.run()
     note_engine_for_statistics(runner)
     if runner.interesting_examples:
